@@ -2,7 +2,7 @@
 
 ;; Copyright (C) 2018 Free Software Foundation, Inc.
 
-;; Version: 0.2
+;; Version: 0.3
 ;; Author: João Távora <joaotavora@gmail.com>
 ;; Maintainer: João Távora <joaotavora@gmail.com>
 ;; URL: https://github.com/joaotavora/eglot
@@ -71,7 +71,9 @@
 (defvar eglot-server-programs '((rust-mode . ("rls"))
                                 (python-mode . ("pyls"))
                                 (js-mode . ("javascript-typescript-stdio"))
-                                (sh-mode . ("bash-language-server" "start")))
+                                (sh-mode . ("bash-language-server" "start"))
+                                (php-mode . ("php" "vendor/felixfbecker/\
+language-server/bin/php-language-server.php")))
   "Alist mapping major modes to server executables.")
 
 (defface eglot-mode-line
@@ -193,6 +195,7 @@ CONTACT is as `eglot--contact'.  Returns a process object."
             (make-process :name readable-name
                           :buffer buffer
                           :command contact
+                          :coding 'no-conversion
                           :connection-type 'pipe
                           :stderr (get-buffer-create (format "*%s stderr*"
                                                              name))))))
@@ -276,6 +279,7 @@ INTERACTIVE is t if inside interactive call."
                                                     'network)
                                           (emacs-pid))
                              :capabilities(eglot--client-capabilities)
+                             :rootPath  (car (project-roots project))
                              :rootUri  (eglot--path-to-uri
                                         (car (project-roots project)))
                              :initializationOptions  []))
@@ -538,7 +542,7 @@ is a symbol saying if this is a client or server originated."
 
 (defun eglot--process-receive (proc message)
   "Process MESSAGE from PROC."
-  (cl-destructuring-bind (&key method id error &allow-other-keys) message
+  (cl-destructuring-bind (&key method id params error result _jsonrpc) message
     (let* ((continuations (and id
                                (not method)
                                (gethash id (eglot--pending-continuations proc)))))
@@ -548,24 +552,19 @@ is a symbol saying if this is a client or server originated."
              ;; a server notification or a server request
              (let* ((handler-sym (intern (concat "eglot--server-" method))))
                (if (functionp handler-sym)
-                   (apply handler-sym proc (append
-                                            (plist-get message :params)
-                                            (if id `(:id ,id))))
+                   ;; FIXME: will fail if params is array instead of  not an object
+                   (apply handler-sym proc (append params (if id `(:id ,id))))
                  (eglot--warn "No implementation of method %s yet" method)
                  (when id
                    (eglot--reply
                     proc id
-                    :error (eglot--obj :code -32601
-                                       :message "Method unimplemented"))))))
+                    :error `(:code -32601 :message "Method unimplemented"))))))
             (continuations
              (cancel-timer (cl-third continuations))
              (remhash id (eglot--pending-continuations proc))
              (if error
-                 (apply (cl-second continuations) error)
-               (let ((res (plist-get message :result)))
-                 (if (listp res)
-                     (apply (cl-first continuations) res)
-                   (funcall (cl-first continuations) res)))))
+                 (funcall (cl-second continuations) error)
+               (funcall (cl-first continuations) result)))
             (id
              (eglot--warn "Ooops no continuation for id %s" id)))
       (eglot--call-deferred proc)
@@ -613,7 +612,8 @@ request request and a process object.")
 
 (cl-defmacro eglot--lambda (cl-lambda-list &body body)
   (declare (indent 1) (debug (sexp &rest form)))
-  `(cl-function (lambda ,cl-lambda-list ,@body)))
+  (let ((e (gensym "eglot--lambda-elem")))
+    `(lambda (,e) (apply (cl-function (lambda ,cl-lambda-list ,@body)) ,e))))
 
 (cl-defun eglot--async-request (proc
                                 method
@@ -622,12 +622,14 @@ request request and a process object.")
                                 &key success-fn error-fn timeout-fn
                                 (timeout eglot-request-timeout)
                                 (deferred nil))
-  "Make a request to PROCESS, expecting a reply.
-Return the ID of this request. Wait TIMEOUT seconds for response.
-If DEFERRED, maybe defer request to the future, or never at all,
-in case a new request with identical DEFERRED and for the same
-buffer overrides it. However, if that happens, the original
-timeout keeps counting."
+  "Make a request to PROCESS, expecting a reply later on.
+SUCCESS-FN and ERROR-FN are passed `:result' and `:error'
+objects, respectively.  Wait TIMEOUT seconds for response or call
+nullary TIMEOUT-FN.  If DEFERRED, maybe defer request to the
+future, or to never at all, in case a new request with identical
+DEFERRED and for the same buffer overrides it (however, if that
+happens, the original timeout keeps counting). Return the ID of
+this request."
   (let* ((id (eglot--next-request-id))
          (existing-timer nil)
          (make-timeout
@@ -689,23 +691,18 @@ DEFERRED is passed to `eglot--async-request', which see."
   (when deferred (eglot--signal-textDocument/didChange))
   (let* ((done (make-symbol "eglot--request-catch-tag"))
          (res
-          (catch done (eglot--async-request
-                       proc method params
-                       :success-fn (lambda (&rest args)
-                                     (throw done (if (vectorp (car args))
-                                                     (car args) args)))
-                       :error-fn (eglot--lambda
-                                     (&key code message &allow-other-keys)
-                                   (throw done
-                                          `(error ,(format "Oops: %s: %s"
-                                                           code message))))
-                       :timeout-fn (lambda ()
-                                     (throw done '(error "Timed out")))
-                       :deferred deferred)
-                 ;; now spin, baby!
-                 (while t (accept-process-output nil 0.01)))))
-    (when (and (listp res) (eq 'error (car res))) (eglot--error (cadr res)))
-    res))
+          (catch done
+            (eglot--async-request
+             proc method params
+             :success-fn (lambda (result) (throw done `(done ,result)))
+             :timeout-fn (lambda () (throw done '(error "Timed out")))
+             :error-fn (eglot--lambda (&key code message _data)
+                         (throw done `(error
+                                       ,(format "Ooops: %s: %s" code message))))
+             :deferred deferred)
+            (while t (accept-process-output nil 30)))))
+    (when (eq 'error (car res)) (eglot--error (cadr res)))
+    (cadr res)))
 
 (cl-defun eglot--notify (process method params)
   "Notify PROCESS of something, don't expect a reply.e"
@@ -725,11 +722,11 @@ DEFERRED is passed to `eglot--async-request', which see."
 ;;;
 (defun eglot--error (format &rest args)
   "Error out with FORMAT with ARGS."
-  (error (apply #'format format args)))
+  (error "[eglot] %s" (apply #'format format args)))
 
 (defun eglot--message (format &rest args)
   "Message out with FORMAT with ARGS."
-  (message (concat "[eglot] " (apply #'format format args))))
+  (message "[eglot] %s" (apply #'format format args)))
 
 (defun eglot--warn (format &rest args)
   "Warning message with FORMAT and ARGS."
@@ -759,20 +756,17 @@ DEFERRED is passed to `eglot--async-request', which see."
                            (line-beginning-position))))
                   (point)))
 
-
-(defun eglot--mapply (fun seq)
-  "Apply FUN to every element of SEQ."
-  (mapcar (lambda (e) (apply fun e)) seq))
-
 (defun eglot--path-to-uri (path)
-  "Urify PATH."
-  (url-hexify-string (concat "file://" (file-truename path))
-                     url-path-allowed-chars))
+  "URIfy PATH."
+  (url-hexify-string
+   (concat "file://" (if (eq system-type 'windows-nt) "/") (file-truename path))
+   url-path-allowed-chars))
 
 (defun eglot--uri-to-path (uri)
   "Convert URI to a file path."
   (when (keywordp uri) (setq uri (substring (symbol-name uri) 1)))
-  (url-filename (url-generic-parse-url (url-unhex-string uri))))
+  (let ((retval (url-filename (url-generic-parse-url (url-unhex-string uri)))))
+    (if (eq system-type 'windows-nt) (substring retval 1) retval)))
 
 (defconst eglot--kind-names
   `((1 . "Text") (2 . "Method") (3 . "Function") (4 . "Constructor")
@@ -802,14 +796,10 @@ DEFERRED is passed to `eglot--async-request', which see."
   "Determine if current server is capable of FEAT."
   (plist-get (eglot--capabilities (eglot--current-process-or-lose)) feat))
 
-(cl-defmacro eglot--with-lsp-range ((start end) range &body body
-                                    &aux (range-sym (cl-gensym)))
-  "Bind LSP RANGE to START and END. Evaluate BODY."
-  (declare (indent 2) (debug (sexp sexp &rest form)))
-  `(let* ((,range-sym ,range)
-          (,start (eglot--lsp-position-to-point (plist-get ,range-sym :start)))
-          (,end (eglot--lsp-position-to-point (plist-get ,range-sym :end))))
-     ,@body))
+(defun eglot--range-region (range)
+  "Return region (BEG . END) that represents LSP RANGE."
+  (cons (eglot--lsp-position-to-point (plist-get range :start))
+        (eglot--lsp-position-to-point (plist-get range :end))))
 
 
 ;;; Minor modes
@@ -1002,18 +992,14 @@ called interactively."
 (cl-defun eglot--server-textDocument/publishDiagnostics
     (_process &key uri diagnostics)
   "Handle notification publishDiagnostics"
-  (let* ((obj (url-generic-parse-url uri))
-	 (filename (car (url-path-and-query obj)))
-         (buffer (find-buffer-visiting filename)))
-    (cond
-     (buffer
+  (if-let ((buffer (find-buffer-visiting (eglot--uri-to-path uri))))
       (with-current-buffer buffer
         (cl-loop
          for diag-spec across diagnostics
          collect (cl-destructuring-bind (&key range severity _group
                                               _code source message)
                      diag-spec
-                   (eglot--with-lsp-range (beg end) range
+                   (pcase-let ((`(,beg . ,end) (eglot--range-region range)))
                      (flymake-make-diagnostic (current-buffer)
                                               beg end
                                               (cond ((<= severity 1) :error)
@@ -1025,9 +1011,8 @@ called interactively."
                         (funcall eglot--current-flymake-report-fn diags)
                         (setq eglot--unreported-diagnostics nil))
                        (t
-                        (setq eglot--unreported-diagnostics diags))))))
-     (t
-      (eglot--message "OK so %s isn't visited" filename)))))
+                        (setq eglot--unreported-diagnostics diags)))))
+    (eglot--warn "Diagnostics received for unvisited %s" uri)))
 
 (cl-defun eglot--register-unregister (proc jsonrpc-id things how)
   "Helper for `eglot--server-client/registerCapability'.
@@ -1233,7 +1218,7 @@ DUMMY is ignored"
       (completion-table-with-cache
        (lambda (string)
          (setq eglot--xref-known-symbols
-               (eglot--mapply
+               (mapcar
                 (eglot--lambda (&key name kind location containerName)
                   (propertize name
                               :textDocumentPositionParams
@@ -1266,10 +1251,9 @@ DUMMY is ignored"
                             :textDocument/definition
                             (get-text-property
                              0 :textDocumentPositionParams identifier)))))
-    (eglot--mapply
-     (eglot--lambda (&key uri range)
-       (eglot--xref-make identifier uri (plist-get range :start)))
-     location-or-locations)))
+    (mapcar (eglot--lambda (&key uri range)
+              (eglot--xref-make identifier uri (plist-get range :start)))
+            location-or-locations)))
 
 (cl-defmethod xref-backend-references ((_backend (eql eglot)) identifier)
   (unless (eglot--server-capable :referencesProvider)
@@ -1280,25 +1264,23 @@ DUMMY is ignored"
                (and rich (get-text-property 0 :textDocumentPositionParams rich))))))
     (unless params
       (eglot--error "Don' know where %s is in the workspace!" identifier))
-    (eglot--mapply
-     (eglot--lambda (&key uri range)
-       (eglot--xref-make identifier uri (plist-get range :start)))
-     (eglot--request (eglot--current-process-or-lose)
-                     :textDocument/references
-                     (append
-                      params
-                      (eglot--obj :context
-                                  (eglot--obj :includeDeclaration t)))))))
+    (mapcar (eglot--lambda (&key uri range)
+              (eglot--xref-make identifier uri (plist-get range :start)))
+            (eglot--request (eglot--current-process-or-lose)
+                            :textDocument/references
+                            (append
+                             params
+                             (eglot--obj :context
+                                         (eglot--obj :includeDeclaration t)))))))
 
 (cl-defmethod xref-backend-apropos ((_backend (eql eglot)) pattern)
   (when (eglot--server-capable :workspaceSymbolProvider)
-    (eglot--mapply
-     (eglot--lambda (&key name location &allow-other-keys)
-       (cl-destructuring-bind (&key uri range) location
-         (eglot--xref-make name uri (plist-get range :start))))
-     (eglot--request (eglot--current-process-or-lose)
-                     :workspace/symbol
-                     (eglot--obj :query pattern)))))
+    (mapcar (eglot--lambda (&key name location &allow-other-keys)
+              (cl-destructuring-bind (&key uri range) location
+                (eglot--xref-make name uri (plist-get range :start))))
+            (eglot--request (eglot--current-process-or-lose)
+                            :workspace/symbol
+                            (eglot--obj :query pattern)))))
 
 (defun eglot-completion-at-point ()
   "EGLOT's `completion-at-point' function."
@@ -1315,7 +1297,7 @@ DUMMY is ignored"
                                        (eglot--TextDocumentPositionParams)
                                        :textDocument/completion))
                  (items (if (vectorp resp) resp (plist-get resp :items))))
-            (eglot--mapply
+            (mapcar
              (eglot--lambda (&rest all &key label &allow-other-keys)
                (add-text-properties 0 1 all label) label)
              items))))
@@ -1352,7 +1334,7 @@ DUMMY is ignored"
 
 (defun eglot--hover-info (contents &optional range)
   (concat (and range
-               (eglot--with-lsp-range (beg end) range
+               (pcase-let ((`(,beg . ,end) (eglot--range-region range)))
                  (concat (buffer-substring beg end)  ": ")))
           (mapconcat #'eglot--format-markup
                      (append
@@ -1431,14 +1413,14 @@ If SKIP-SIGNATURE, don't try to send textDocument/signatureHelp."
                        (mapc #'delete-overlay eglot--highlights)
                        (setq eglot--highlights
                              (when-buffer-window
-                              (eglot--mapply
-                               (eglot--lambda (&key range _kind)
-                                 (eglot--with-lsp-range (beg end) range
-                                   (let ((ov (make-overlay beg end)))
-                                     (overlay-put ov 'face 'highlight)
-                                     (overlay-put ov 'evaporate t)
-                                     ov)))
-                               highlights))))
+                              (mapcar (eglot--lambda (&key range _kind)
+                                        (pcase-let ((`(,beg . ,end)
+                                                     (eglot--range-region range)))
+                                          (let ((ov (make-overlay beg end)))
+                                            (overlay-put ov 'face 'highlight)
+                                            (overlay-put ov 'evaporate t)
+                                            ov)))
+                                      highlights))))
          :deferred :textDocument/documentHighlight))))
   nil)
 
@@ -1446,7 +1428,7 @@ If SKIP-SIGNATURE, don't try to send textDocument/signatureHelp."
   "EGLOT's `imenu-create-index-function' overriding OLDFUN."
   (if (eglot--server-capable :documentSymbolProvider)
       (let ((entries
-             (eglot--mapply
+             (mapcar
               (eglot--lambda (&key name kind location _containerName)
                 (cons (propertize name :kind (cdr (assoc kind eglot--kind-names)))
                       (eglot--lsp-position-to-point
@@ -1466,14 +1448,13 @@ If SKIP-SIGNATURE, don't try to send textDocument/signatureHelp."
   (unless (or (not version) (equal version eglot--versioned-identifier))
     (eglot--error "Edits on `%s' require version %d, you have %d"
                   (current-buffer) version eglot--versioned-identifier))
-  (eglot--mapply
-   (eglot--lambda (&key range newText)
-     (save-restriction
-       (widen)
-       (save-excursion
-         (eglot--with-lsp-range (beg end) range
-           (goto-char beg) (delete-region beg end) (insert newText)))))
-   edits)
+  (mapc (eglot--lambda (&key range newText)
+          (save-restriction
+            (widen)
+            (save-excursion
+              (pcase-let ((`(,beg . ,end) (eglot--range-region range)))
+                (goto-char beg) (delete-region beg end) (insert newText)))))
+        edits)
   (eglot--message "%s: Performed %s edits" (current-buffer) (length edits)))
 
 (defun eglot--apply-workspace-edit (wedit &optional confirm)
